"""

OpenNEM v2 Energy Methods

"""
from datetime import date, datetime, timedelta
from typing import List, Tuple

import pandas as pd

GeneratedEnergyForDuid = Tuple[date, str, float]


def __trapezium_integration(d_ti: pd.DataFrame, power_column: str = "INITIALMW") -> float:
    """ Trapezium integration fixed at 6 intervals and 30 minutes """
    if power_column not in d_ti:
        raise Exception("No column {}".format(power_column))

    return 0.5 * (d_ti[power_column] * [1, 2, 2, 2, 2, 2, 1]).sum() / 12


def __trading_energy_generator(
    df: pd.DataFrame, duid_id: str, sel_day: date
) -> List[GeneratedEnergyForDuid]:
    df.sort_index(inplace=True)
    t_start = datetime(sel_day.year, sel_day.month, sel_day.day, 0, 5)
    generated_values = []

    # 48 trading intervals in the day
    # (could be better with groupby function)
    for TI in range(48):
        t_i = t_start + timedelta(0, 1800 * TI)
        t_f = t_start + timedelta(0, 1800 * (TI + 1))

        d_ti = df[(df.index >= t_i) & (df.index <= t_f) & (df.DUID == duid_id)]

        if not d_ti.index.empty:
            generated_values.append((d_ti.index[-2], duid_id, __trapezium_integration(d_ti)))

    return generated_values


def trading_energy_data(df: pd.DataFrame, duids: List[str], sel_day: date) -> pd.DataFrame:
    energy_genrecs = []

    for duid in duids:
        energy_genrecs += [d for d in __trading_energy_generator(df, duid, sel_day)]

    df = pd.DataFrame(energy_genrecs, columns=["SETTLEMENTDATE", "DUID", "OUTPUT_MWH"])

    return df



"""
    Create stations on NEM for import/export series
    and facility for one each of import/export
"""

import logging
from datetime import datetime

from opennem.core.dispatch_type import DispatchType
from opennem.core.flows import FlowDirection, generated_flow_station_id
from opennem.core.networks import get_network_region_schema, state_from_network_region
from opennem.db import SessionLocal
from opennem.db.models.opennem import Facility, Location, Station
from opennem.schema.network import NetworkNEM, NetworkSchema

logger = logging.getLogger("opennem.importer.trading_flows")

IMPORTER_NAME = "opennem.importer.trading_flows"


def setup_network_flow_stations(network: NetworkSchema = NetworkNEM) -> None:
    """Creats a station for each network region and a facility for each of imports
    exports so that energy values in facility_scada can be matched up per-region"""

    session = SessionLocal()

    network_regions = get_network_region_schema(network)

    for network_region in network_regions:
        logger.info(f"Setting up for network {network.code} and region: {network_region.code}")

        flow_station_id = generated_flow_station_id(network, network_region)

        flow_station = (
            session.query(Station)
            .filter_by(code=flow_station_id)
            .filter_by(network_code=flow_station_id)
            .one_or_none()
        )

        if not flow_station:
            flow_station = Station(code=flow_station_id, network_code=flow_station_id)

        flow_station.approved = True
        flow_station.approved_by = IMPORTER_NAME
        flow_station.created_by = IMPORTER_NAME

        if not flow_station.location:
            flow_station.location = Location(state=state_from_network_region(network_region.code))

        flow_station.name = "Flows for {} state {}".format(
            network.code.upper(), state_from_network_region(network_region.code.upper())
        )

        flow_facilities = [
            (i, generated_flow_station_id(network, network_region, i)) for i in FlowDirection
        ]

        for (flow_direction, flow_facility_id) in flow_facilities:
            flow_facility_model = (
                session.query(Facility)
                .filter_by(code=flow_facility_id)
                .filter_by(dispatch_type=DispatchType.GENERATOR)
                .filter_by(network_id=network.code)
                .filter_by(network_region=network_region.code)
                .one_or_none()
            )

            if not flow_facility_model:
                flow_facility_model = Facility(  # type: ignore
                    code=flow_facility_id,
                    dispatch_type=DispatchType.GENERATOR,
                    network_id=network.code,
                    network_region=network_region.code,
                )

            flow_facility_model.status_id = "operating"
            flow_facility_model.approved = True
            flow_facility_model.approved_by = IMPORTER_NAME
            flow_facility_model.created_by = IMPORTER_NAME
            flow_facility_model.fueltech_id = flow_direction.value
            flow_facility_model.updated_at = datetime.now()

            flow_station.facilities.append(flow_facility_model)

            session.add(flow_station)

            logger.info(
                "Created network trading flow station facility: {}".format(flow_facility_id)
            )

    session.commit()

    return None


if __name__ == "__main__":
    setup_network_flow_stations()



from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session
from starlette import status

from opennem.core.dispatch_type import DispatchType
from opennem.db import get_database_session
from opennem.db.models.opennem import Facility, FuelTech, Location, Network, Revision, Station
from opennem.schema.opennem import StationSchema

from .schema import StationIDList, StationRecord

router = APIRouter()


@router.get(
    "/",
    response_model=List[StationRecord],
    description="Get a list of all stations",
    response_model_exclude_none=True,
)
def stations(
    session: Session = Depends(get_database_session),
    facilities_include: Optional[bool] = Query(False, description="Include facilities in records"),
    revisions_include: Optional[bool] = Query(False, description="Include revisions in records"),
    history_include: Optional[bool] = Query(False, description="Include history in records"),
    only_approved: Optional[bool] = Query(
        False, description="Only show approved stations not those pending"
    ),
    name: Optional[str] = None,
    limit: Optional[int] = None,
    page: int = 1,
) -> List[StationSchema]:
    stations = (
        session.query(Station)
        .join(Location)
        .enable_eagerloads(True)
        #
        # .group_by(Station.code)
        # .filter(Facility.fueltech_id.isnot(None))
        # .filter(Facility.status_id.isnot(None))
    )

    if facilities_include:
        stations = stations.outerjoin(Facility, Facility.station_id == Station.id).outerjoin(
            FuelTech, Facility.fueltech_id == FuelTech.code
        )

    if revisions_include:
        stations = stations.outerjoin(Revision, Revision.station_id == Station.id)

    if only_approved:
        stations = stations.filter(Station.approved == True)

    if name:
        stations = stations.filter(Station.name.like("%{}%".format(name)))

    stations = stations.order_by(
        # Facility.network_region,
        Station.name,
        # Facility.network_code,
        # Facility.code,
    )

    stations = stations.all()

    return stations


@router.get(
    "/ids",
    response_model=List[StationIDList],
    description="Get a list of station ids for dropdowns",
    response_model_exclude_none=True,
)
def station_ids(
    session: Session = Depends(get_database_session),
    only_approved: Optional[bool] = Query(
        True, description="Only show approved stations not those pending"
    ),
) -> List[StationIDList]:
    stations = session.query(Station).join(Station.location)

    if only_approved:
        stations = stations.filter(Station.approved == True)

    stations = stations.order_by(
        Station.id,
    )

    stations = stations.all()

    return stations


@router.get(
    "/{country_code}/{network_id}/{station_code:path}",
    response_model=StationSchema,
    description="Get a single station by code",
    response_model_exclude_none=True,
)
def station(
    country_code: str,
    network_id: str,
    station_code: str,
    session: Session = Depends(get_database_session),
    only_generators: bool = Query(True, description="Show only generators"),
) -> StationSchema:

    station = (
        session.query(Station)
        .filter(Station.code == station_code)
        .filter(Facility.station_id == Station.id)
        .filter(~Facility.code.endswith("NL1"))
        .filter(Facility.network_id == network_id)
        .filter(Network.country == country_code)
    )

    if only_generators:
        station = station.filter(Facility.dispatch_type == DispatchType.GENERATOR)

    station = station.one_or_none()

    if not station:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Station not found")

    if not station.facilities:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Station has no facilities",
        )

    station.network = station.facilities[0].network_id

    return station



from hashlib import md5
from io import BytesIO

from PIL import Image


def img_to_buffer(img: Image) -> memoryview:
    """
    Convert image and save as JPEG

    @TODO jpeg settings in opennem.settings

    """
    buf = BytesIO()

    # convert all to RGP
    if img.mode in ("RGBA", "P", "LA"):
        img = img.convert("RGB")

    img.save(buf, format="JPEG")

    return buf.getbuffer()


def image_get_hash(img: Image) -> str:
    """Image hash based on content"""
    img = img.resize((10, 10), Image.ANTIALIAS)
    img = img.convert("L")

    pixel_data = list(img.getdata())
    avg_pixel = sum(pixel_data) / len(pixel_data)
    bits = "".join(["1" if (px >= avg_pixel) else "0" for px in pixel_data])
    hex_representation = str(hex(int(bits, 2)))[2:][::-1].upper()

    return hex_representation.lower()


def image_get_crypto_hash(img: Image, save_driver: str = "JPEG") -> str:
    """Get a cryptographic hash of an image"""
    img_hash = md5()

    # convert to RGB
    img_rgb = img.convert("RGB")

    with BytesIO() as memobj:
        img_rgb.save(memobj, save_driver)
        data = memobj.getvalue()
        img_hash.update(data)

    return img_hash.hexdigest()



# pylint: disable=no-member
"""
net interchange

Revision ID: 669bfb3b1245
Revises: 3afcf7bf755c
Create Date: 2021-03-05 04:29:17.113967

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "669bfb3b1245"
down_revision = "3afcf7bf755c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "balancing_summary",
        sa.Column("net_interchange_trading", sa.Numeric(), nullable=True),
    )


def downgrade() -> None:
    op.drop_column("balancing_summary", "net_interchange_trading")
